#include "segway_RMP400.h"
#include "segway_RMP400_exception.h"
#include "segway_rmp200_exceptions.h"
#include <cmath>

CSegwayRMP400::CSegwayRMP400() :
  forward_displacement(0.0),
  yaw_displacement(0.0),
  yaw_rate(0.0)
{
  this->connected=false;
  this->segways.resize(NUM_SEGWAY_200);
}

CSegwayRMP400::CSegwayRMP400(const std::string &front, const std::string &rear) :
  forward_displacement(0.0),
  yaw_displacement(0.0),
  yaw_rate(0.0)
{
  this->connected=false;
  this->segways.resize(NUM_SEGWAY_200);
  this->connect(front,rear);
}

void *CSegwayRMP400::monitor_thread(void *param)
{
  CSegwayRMP400 *segway=(CSegwayRMP400 *)param;
  std::vector<bool> status_received(false,NUM_SEGWAY_200);
  double fwd_displ_front,fwd_displ_rear; 
  std::list<std::string> events;
  int event_id=0;
  bool end=false;

  events.push_back(segway->finish_thread_event_id);
  events.push_back(segway->segways[front_segway].get_new_status_event());
  events.push_back(segway->segways[rear_segway].get_new_status_event());
  while(!end)
  {
    event_id=segway->event_server->wait_first(events);
    segway->status_access.enter();
    try{
      if(event_id==0)
        end=true;
      else if(event_id==1)// front status
      {
        status_received[front_segway]=true;
        segway->status.rmp200[front_segway]=segway->segways[front_segway].get_status();
      }
      else if(event_id==2)// rear status
      {
        status_received[rear_segway]=true;
        segway->status.rmp200[rear_segway]=segway->segways[rear_segway].get_status();
      }
      if(status_received[front_segway] && status_received[rear_segway])
      {
        /* compute the forward displacement  */
        fwd_displ_front=(segway->segways[front_segway].get_right_wheel_displacement()+segway->segways[front_segway].get_left_wheel_displacement())/2;
        fwd_displ_rear=(segway->segways[rear_segway].get_right_wheel_displacement()+segway->segways[rear_segway].get_left_wheel_displacement())/2;
        segway->forward_displacement=(fwd_displ_front+fwd_displ_rear)/2;
        /* compute yaw displacement*/
        segway->yaw_displacement=(segway->segways[front_segway].get_yaw_displacement()+segway->segways[rear_segway].get_yaw_displacement())/2;
        /* compute yaw rate */
        segway->yaw_rate=(segway->segways[front_segway].get_yaw_rate()+segway->segways[rear_segway].get_yaw_rate())/2;
        /* activate event */
        if(!segway->event_server->event_is_set(segway->new_status_event_id))
          segway->event_server->set_event(segway->new_status_event_id);
        status_received[front_segway]=false;
        status_received[rear_segway]=false;
      }
    }catch(CException &e){
      std::cout << e.what() << std::endl;
    }
    segway->status_access.exit();
  }

  pthread_exit(NULL);
}

void CSegwayRMP400::init_events_threads(void)
{
  std::string id="segway_"+this->front_serial+"_"+this->rear_serial;

  /* initialize threads */
  this->thread_server=CThreadServer::instance();
  this->monitor_thread_id=id+"_monitor_thread";
  this->thread_server->create_thread(this->monitor_thread_id);
  this->thread_server->attach_thread(this->monitor_thread_id,this->monitor_thread,this);
  /* initialize events */
  this->event_server=CEventServer::instance();
  this->finish_thread_event_id=id+"_finish_thread_event";
  this->event_server->create_event(this->finish_thread_event_id);
  this->front_cable_disconnect_event_id=id+"_front_disconnect_event";
  this->event_server->create_event(this->front_cable_disconnect_event_id);
  this->rear_cable_disconnect_event_id=id+"_rear_disconnect_event";
  this->event_server->create_event(this->rear_cable_disconnect_event_id);
  this->front_power_off_event_id=id+"_front_power_off_event";
  this->event_server->create_event(this->front_power_off_event_id);
  this->rear_power_off_event_id=id+"_rear_power_off_event";
  this->event_server->create_event(this->rear_power_off_event_id);
  this->front_no_heartbeat_event_id=id+"_front_no_heartbeat_event";
  this->event_server->create_event(this->front_no_heartbeat_event_id);
  this->rear_no_heartbeat_event_id=id+"_rear_no_heartbeat_event";
  this->event_server->create_event(this->rear_no_heartbeat_event_id);
  this->new_status_event_id=id+"_new_status_event";
  this->event_server->create_event(this->new_status_event_id);
  // start the internal thread
  this->thread_server->start_thread(this->monitor_thread_id);
}

void CSegwayRMP400::connect(const std::string &front, const std::string &rear)
{
  if(!this->connected)
  {
    if(front==rear)
      throw CSegwayRMP400Exception(_HERE_,"Rear and front serial number can't be the same",front);
    else
    {
      try{
        this->segways[front_segway].connect(front);
        this->front_serial=front;
        this->segways[rear_segway].connect(rear);
        this->rear_serial=rear;
        this->init_events_threads();
        this->connected=true;
      }catch(CException &e){
        this->close();
        throw e;
      }
    }
  }
  else
    throw CSegwayRMP400Exception(_HERE_,"The driver is already connected",front+"_"+rear);
}

void CSegwayRMP400::close()
{
  // signal the thread to end
  if(this->monitor_thread_id.size()>0)
  {
    if(this->thread_server->get_thread_state(this->monitor_thread_id)==starting ||
       this->thread_server->get_thread_state(this->monitor_thread_id)==active)
    {
      this->event_server->set_event(this->finish_thread_event_id);
      this->thread_server->end_thread(this->monitor_thread_id);
      this->thread_server->detach_thread(this->monitor_thread_id);
      this->thread_server->delete_thread(this->monitor_thread_id);
      this->monitor_thread_id.clear();
      this->event_server->delete_event(this->finish_thread_event_id);
      this->finish_thread_event_id.clear();
      this->event_server->delete_event(this->front_cable_disconnect_event_id);
      this->front_cable_disconnect_event_id.clear();
      this->event_server->delete_event(this->rear_cable_disconnect_event_id);
      this->rear_cable_disconnect_event_id.clear();
      this->event_server->delete_event(this->front_power_off_event_id);
      this->front_power_off_event_id.clear();
      this->event_server->delete_event(this->rear_power_off_event_id);
      this->rear_power_off_event_id.clear();
      this->event_server->delete_event(this->front_no_heartbeat_event_id);
      this->front_no_heartbeat_event_id.clear();
      this->event_server->delete_event(this->rear_no_heartbeat_event_id);
      this->rear_no_heartbeat_event_id.clear();
      this->event_server->delete_event(this->new_status_event_id);
      this->new_status_event_id.clear();
    }
  }
  if(this->front_serial.size()>0) 
  {
    this->segways[front_segway].close();
    this->front_serial.clear();
  }
  if(this->rear_serial.size()>0)
  {
    this->segways[rear_segway].close();
    this->rear_serial.clear();
  }
  this->connected=false;
}

std::ostream & operator<< (std::ostream& data, CSegwayRMP400 & segway)
{
  data << "Front segway status data:" << std::endl;
  data << segway.segways[front_segway] << std::endl;
  data << "Rear segway status data:" << std::endl;
  data << segway.segways[rear_segway] << std::endl;

  return data;
}

TSegwayRMP400Status CSegwayRMP400::get_status()
{
  TSegwayRMP400Status status_tmp;

  if(this->connected)
  {
    this->status_access.enter();
    status_tmp=this->status;
    this->status_access.exit();

    return status_tmp;
  }
  else
    throw CSegwayRMP400Exception(_HERE_,"The driver is not connected","");
}

double CSegwayRMP400::get_forward_displacement()
{
  double fwd_disp;

  if(this->connected)
  {
    this->status_access.enter();
    fwd_disp=this->forward_displacement;
    this->status_access.exit();

    return fwd_disp;
  }
  else
    throw CSegwayRMP400Exception(_HERE_,"The driver is not connected","");
}

double CSegwayRMP400::get_yaw_displacement()
{
  double yaw_disp;

  if(this->connected)
  {
    this->status_access.enter();
    yaw_disp=this->yaw_displacement;
    this->status_access.exit();

    return yaw_disp;
  }
  else
    throw CSegwayRMP400Exception(_HERE_,"The driver is not connected","");
}

double CSegwayRMP400::get_yaw_rate()
{
  double yaw_rate_tmp;

  if(this->connected)
  {
    this->status_access.enter();
    yaw_rate_tmp=this->yaw_rate;
    this->status_access.exit();

    return yaw_rate_tmp;
  }
  else
    throw CSegwayRMP400Exception(_HERE_,"The driver is not connected","");
}

void CSegwayRMP400::reset_integrators()
{
  double fd=0.0,yd=0.0;
  unsigned int i=0;
  bool reset=false;

  if(this->connected)
  {
    do {
      for(i=0;i<this->segways.size();i++)
      {
        this->segways[i].reset_right_wheel_integrator();
        this->segways[i].reset_left_wheel_integrator();
        this->segways[i].reset_yaw_integrator();
        this->segways[i].reset_forward_integrator();
      }
      // Need to leave the segwayrmp 200 driver to read need values
      sleep(1);
      fd=this->get_forward_displacement();
      yd=this->get_yaw_displacement();
      if(std::fabs(fd)<0.1 && std::fabs(yd)<0.1)
        reset=true;
      else
        reset=false;
    }while(!reset);
  }
  else
    throw CSegwayRMP400Exception(_HERE_,"The driver is not connected","");
}

std::string CSegwayRMP400::get_front_cable_disconnected_event_id(void)
{
  return this->front_cable_disconnect_event_id;
}

std::string CSegwayRMP400::get_rear_cable_disconnected_event_id(void)
{
  return this->rear_cable_disconnect_event_id;
}

std::string CSegwayRMP400::get_front_power_off_event_id(void)
{
  return this->front_power_off_event_id;
}

std::string CSegwayRMP400::get_rear_power_off_event_id(void)
{
  return this->rear_power_off_event_id;
}

std::string CSegwayRMP400::get_front_no_heartbeat_event_id(void)
{
  return this->front_no_heartbeat_event_id;
}

std::string CSegwayRMP400::get_rear_no_heartbeat_event_id(void)
{
  return this->rear_no_heartbeat_event_id;
}

std::string CSegwayRMP400::get_new_status_event_id(void)
{
  return this->new_status_event_id;
}

void CSegwayRMP400::move(double vT, double vR)
{
  if(this->connected)
  {
    this->segways[front_segway].move(vT,vR);
    this->segways[rear_segway].move(vT,vR);
  }
  else
    throw CSegwayRMP400Exception(_HERE_,"The driver is not connected","");
}

void CSegwayRMP400::stop()
{
  if(this->connected)
  {
    this->segways[front_segway].stop();
    this->segways[rear_segway].stop();
  }
  else
    throw CSegwayRMP400Exception(_HERE_,"The driver is not connected","");
}

CSegwayRMP400::~CSegwayRMP400()
{
  this->close();
}
